#ifndef PROTOCOLMANAGER_H
#define	PROTOCOLMANAGER_H

#include <Response.pb.h>
#include <Request.pb.h>
#include <Configuration.h>
#include <DBManager.h>

#include <tango.h>

#include <boost/date_time.hpp>

namespace DataImporter_ns
{

//Protocol buffer request class shared pointer
typedef boost::shared_ptr<Request> RequestSP;

//Protocol buffer response class shared pointer
typedef boost::shared_ptr<Response> ResponseSP;

class ProtocolManager : public Tango::LogAdapter
{
public:
//------------------------------------------------------------------------------
//  [Public] Shared pointer typedef
//------------------------------------------------------------------------------
    typedef boost::shared_ptr<ProtocolManager> SP;

protected:
//------------------------------------------------------------------------------
//  [Protected] Constructor destructor deleter
//------------------------------------------------------------------------------
    ProtocolManager(Tango::DeviceImpl*, Configuration::SP, DBManager::SP);

    virtual ~ProtocolManager();

    class Deleter;
    friend Deleter;
    class Deleter
    {
        public:
            void operator()(ProtocolManager* d) { delete d; }
    };

public:
//------------------------------------------------------------------------------
//	[Public] Class creation method
//------------------------------------------------------------------------------
    static ProtocolManager::SP create(Tango::DeviceImpl*, Configuration::SP,
        DBManager::SP);

//------------------------------------------------------------------------------
//	[Public] Remote endpoint setter method
//------------------------------------------------------------------------------
    virtual void setRemoteEndpoint(std::string);

//------------------------------------------------------------------------------
//  [Public] Reset protocol status method
//------------------------------------------------------------------------------
    virtual void resetProtocolStatus();

//------------------------------------------------------------------------------
//  [Public] Request response management method
//------------------------------------------------------------------------------
    virtual RequestSP createRequest()
        throw(std::runtime_error, std::out_of_range);

    virtual void processResponse(ResponseSP)
        throw(std::runtime_error, std::out_of_range);

//------------------------------------------------------------------------------
//	[Public] File transfer methods
//------------------------------------------------------------------------------
    virtual bool isTransferRequest();

    virtual std::string getFileDir();

    virtual std::string getFileName();

    virtual int getFileSize();

protected:
//------------------------------------------------------------------------------
//  [Protected] Request specific methods
//------------------------------------------------------------------------------
    virtual RequestSP createAuthroisation()
        throw(std::runtime_error);

    virtual RequestSP createValidation()
        throw(std::runtime_error);

    virtual RequestSP createTransfer()
        throw(std::runtime_error, std::out_of_range);

    virtual RequestSP createKeepAlive();

//------------------------------------------------------------------------------
//  [Protected] Response specific methods
//------------------------------------------------------------------------------
    virtual void processAuthroisation(ResponseSP)
        throw(std::runtime_error);

    virtual void processValidation(ResponseSP)
        throw(std::runtime_error);

    virtual void processTransfer(ResponseSP)
        throw(std::runtime_error, std::out_of_range);

    virtual void processKeepAlive(ResponseSP);

//------------------------------------------------------------------------------
//  [Protected] Class variables
//------------------------------------------------------------------------------
    //Configuration parameters shared pointer
    Configuration::SP m_configuration_sp;

    //Database manger shared pointer
    DBManager::SP m_dBManager_sp;

    //Client is authorised
    bool m_isAuthorised;

    //Table is validated
    bool m_isValidated;

    //Current timestamp index
    boost::posix_time::ptime m_currentTimestamp;

    //Retrieved data iterator
    DBManager::FileRowset::const_iterator m_it;

    //Retrieved data shared pointer
    DBManager::FileRowsetSP m_rowSet_sp;

    //Address and port of remote endpoint
    std::string m_remoteEndpoint;

    bool m_isTransferRequest;

    std::string m_fileDir;

    std::string m_fileName;

    int m_fileSize;
};

}   //End of namespace

#endif	/* PROTOCOLMANAGER_H */

